// Copyright 2020 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build android || arm64 || riscv64
// +build android arm64 riscv64

package syscall

const _SYS_dup = SYS_DUP3
