// Copyright 2016 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package b

import "./a"

func F() {
	a.F()
	a.Fi()
}

func Fp() {
	a.Fp()
	a.Fip()
}

func Gp() {
	a.Gp()
	a.Gip()
}

func Hp() {
	a.Hp()
	a.Hip()
}
