// Code generated by mkconsts.go. DO NOT EDIT.

//go:build !goexperiment.staticlockranking
// +build !goexperiment.staticlockranking

package goexperiment

const StaticLockRanking = false
const StaticLockRankingInt = 0
