// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build (windows && !amd64) || !windows
// +build windows,!amd64 !windows

package runtime

//go:nosplit
func osSetupTLS(mp *m) {}
