// compiledir

// Copyright 2012 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Exported data for inlining could forget types of
// local variables declared in inlinable bodies.

package ignored
