// Code generated by mkconsts.go. DO NOT EDIT.

//go:build goexperiment.regabireflect
// +build goexperiment.regabireflect

package goexperiment

const RegabiReflect = true
const RegabiReflectInt = 1
